import requests
import time
import gc
import tempfile
import os
import cv2
import statistics
import numpy as np
from . import helpers
from nsfw_detector import predict
from fluxhelper import Logger
from .exceptions import *
from PIL import Image
from io import BytesIO

class Classifier:
    def __init__(self, model: str, logging: Logger = None, **kwargs):
        self.model = predict.load_model(model)

        if not logging:
            self.logging = Logger()
        else:
            self.logging = logging

        # Params
        self.frameUniqueness = kwargs.get('frameUniqueness', 5)
        self.imgSize = kwargs.get("imgSize", (224, 224))
        self.workers = kwargs.get("workers", 4)

    def classify(self, url: str) -> str:
        """
        Classify a image/video url.
        
        Parameters
        ----------
        `content` : str
            The url of the image/video to classify.
        
        Returns
        -------
        {
            "contentType": str,
            "data": {
                if video or gif:
                    "frames": {
                        frame_number: {
                            "sexy": float,
                            "neutral": float,
                            "porn": float,
                            "hentai": float,
                            "drawings": float
                        },
                        ...
                    },
                    "mean": float
                else:
                    "sexy": float,
                    "neutral": float,
                    "porn": float,
                    "hentai": float,
                    "drawings": float
            }
        }

        Raises
        ------
        `InternalRequestError` :
            When there's an error getting the content from the url.
        `UnknownContentType` :
            When the content type is not supported or unknown, mostly raised in video types.
        """
        
        data = {"contentType": None}

        try:
            start = time.time()

            with requests.Session() as session:
                r = session.get(url)

            end = round((time.time() - start) * 1000, 2)
            self.logging.debug(f"Getting content from url took {end}ms")

            r.raise_for_status()
        except requests.exceptions.RequestException as e:
            raise InternalRequestError(e)
        
        data["contentType"] = r.headers["Content-Type"].lower().strip()
        predictionType = "single"

        if data["contentType"] == "image/gif":
            predictionType = "multiple"

            f = BytesIO(r.content)
            frames = helpers.extractGifFrames(f, self.logging, self.frameUniqueness)
            f.close()

        elif data["contentType"].startswith("video/"):
            predictionType = "multiple"
            
            suffix = helpers.VIDEO_MAPPINGS.get(data["contentType"])
            if not suffix:
                raise UnknownContentType(data["contentType"])

            with tempfile.NamedTemporaryFile(delete=False, suffix=suffix) as tmp:
                path = tmp.name
                tmp.write(r.content)
                tmp.close()

                frames = helpers.extractVideoFrames(path, self.logging, self.frameUniqueness, self.workers)
                os.unlink(path)
        elif data["contentType"].startswith("image/"):
            pass 
        else:
            raise UnknownContentType(data["contentType"])
    
        if predictionType == "multiple":
            start = time.time()
            nds = []
            for frame in frames:

                img = frame[0]
                img = img.convert("RGB")

                img = np.array(img, dtype=np.float64)
                img = cv2.resize(img, self.imgSize)
                img /= 255

                nds.append(img)

            nds = np.asarray(nds)
            end = round((time.time() - start) * 1000, 2)
            self.logging.debug(f"Converting frames to numpy arrays and resizing took {end}ms.")

            self.logging.debug(f"Classifying {len(nds)} frames.")
            start = time.time()
            r = predict.classify_nd(self.model, nds)
            _ = gc.collect()
            end = round((time.time() - start) * 1000, 2)
            self.logging.debug(f"Classifying {len(nds)} frames took {end}ms.")
            self.logging.debug(f"Processing {len(r)} results.")

            start = time.time()
            classified = {i: c for i, c in enumerate(r)}
            end = round((time.time() - start) * 1000, 2)
            self.logging.debug(f"Processing {len(r)} results took {end}ms.")

            data["data"] = {
                "frames": classified,
            }

            return data

        elif predictionType == "single":
            img = Image.open(BytesIO(r.content))
            img = img.convert("RGB")
            img = np.array(img, dtype=np.float64)
            img = cv2.resize(img, self.imgSize)
            img /= 255

            start = time.time()
            r = predict.classify_nd(self.model, np.asarray([img]))
            _ = gc.collect()
            end = round((time.time() - start) * 1000, 2)
            self.logging.debug(f"Classifying image took {end}ms.")

            data["data"] = r[0]
            return data



class InternalRequestError(Exception):
    """
    Raised when getting content from a url fails. This could be due to a network error, bad url, bad request, bad status, etc.
    """


class UnknownContentType(Exception):
    """
    Raised when the content type of the response is not known.
    """



from distutils.core import setup
setup(
    name="fastnsfw",
    packages=["fastnsfw"],
    version="0.6",
    license="MIT",
    description="NSFW Detection Library",
    author="Philippe Mathew",
    author_email="philmattdev@gmail.com",
    url="https://github.com/bossauh/fastnsfw",
    download_url="https://github.com/bossauh/fastnsfw/archive/refs/tags/v_06.tar.gz",
    keywords=["nsfw", "detection"],
    install_requires=[
        "fluxhelper",
        "pillow",
        "imagehash",
        "opencv-python"
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7"
    ]
)



from typing import List, Tuple
import cv2
import io
import time
import imagehash
import threading
import cv2
from contextlib import contextmanager
from PIL import Image
from fluxhelper import Logger
from queue import Queue


VIDEO_MAPPINGS = {
    "video/x-msvideo": ".avi",
    "video/mp4": ".mp4",
    "video/mpeg": ".mpeg",
    "video/ogg": ".ogv",
    "video/mp2t": ".ts",
    "video/webm": ".webm",
    "video/x-matroska": ".mkv",
    "video/quicktime": ".mov"
}


@contextmanager
def VideoCapture(*args, **kwargs):
    cap = cv2.VideoCapture(*args, **kwargs)
    try:
        yield cap
    finally:
        cap.release()


def extractGifFrames(filelike: io.BytesIO, logging: Logger, uniqueness: float) -> List[Tuple[Image.Image, imagehash.ImageHash]]:
    """
    Extract frames from a filelike object.
    """

    frames = []

    start = time.time()
    with Image.open(filelike) as img:

        if logging:
            logging.debug(f"Extracting {img.n_frames} frames from a GIF")

        for frame in range(img.n_frames):
            img.seek(frame)
            go = True
            hashed = imagehash.average_hash(img)

            if len(frames) > 1:
                for fc in frames:
                    if (fc[1] - hashed) < uniqueness:
                        go = False
                        break

            if go:
                frames.append((img, hashed))

    end = round((time.time() - start) * 1000, 2)
    if logging:
        logging.debug(f"Extracted {len(frames)} unique frames in {end}ms.")

    return frames


def extractVideoFrames(filepath: str, logging: Logger, uniqueness: float, workers: int) -> List[Tuple[Image.Image, imagehash.ImageHash]]:
    """
    Extract unique frames from a video file path.
    """

    queue = Queue()
    frames = Queue()
    total = 0
    frameCount = cv2.VideoCapture(filepath).get(cv2.CAP_PROP_FRAME_COUNT)

    start = time.time()
    def p():
        while True:
            frame = queue.get()
            if frame is None:
                logging.debug("Exiting worker...")
                break
            
            frame = frame[:, :, ::-1]
            with Image.fromarray(frame) as frame:
                hashed = imagehash.average_hash(frame)
                go = True

                if len(list(frames.queue)) > 1:
                    for fc in list(frames.queue):
                        if (fc[1] - hashed) < uniqueness:
                            go = False
                            break
                if go:
                    frames.put((frame, hashed))
            queue.task_done()

    if frameCount <50:
        logging.warning("There are less than 20 frames, using 1 worker instead as it should do the job.")
        workers = 1

    # Construct workers
    funcs = [p] * workers

    # Launch workers
    for i, t in enumerate(funcs):
        th = threading.Thread(target=t, daemon=True)
        th.start()
        logging.debug(f"Launched worker {i}")
    
    # Put frames into queue
    logging.debug("Extracting frames...")
    with VideoCapture(filepath) as cap:

        status, frame = cap.read()
        while status:
            queue.put(frame)

            total += 1
            status, frame = cap.read()

    # Wait for all workers to be finished
    logging.debug("Waiting for workers to finish...")
    queue.join()

    logging.debug("Workers finished, exiting them...")
    for i in range(workers):
        queue.put(None)

    end = round((time.time() - start) * 1000, 2)
    frames = list(frames.queue)

    if logging:
        logging.debug(
            f"Extracted {len(frames)}/{total} unique frames in {end}ms.")
    return frames
